package Voyager.rtsp.clients;

import org.omg.CORBA.*;
import java.net.*;
import java.util.*;
import java.io.*;

import Voyager.rtsp.*;
import Voyager.util.*;
import Voyager.db.*;

public class RecordController
{
  class Destination
  {
    String host;
    int port;

    Destination(String h, int p)
    {
      host = h;
      port = p;
    }

    //
    // Construct the transport string for an RTSP SETUP
    //
    public String transportString(String mode)
    {
      String portS, hostS, modeS;

      if (host == null)
	{
	  hostS = "";
	}
      else
	{
	  hostS = "destination=" + host + ";";
	}

      portS = "port=" + port + ";";
      modeS = "mode=\"" + mode + "\"";
	    
      return "RTP/AVP;" + hostS + portS + modeS;
    }

    public String toString()
    {
      return host + "/" + port;
    }
  }

  public static final int AUTOMATIC = 1;
  public static final int MANUAL = 2;
  public static final int FLBONE = 4;
  public static final int ACTINIDES = 5;
  String sessionName_;
  String sessionDescription_;

  String[] addresses_;
  int addressMode_;

  Destination audioDestination_;
  Destination videoDestination_;
  Destination destinations_[];

  int sessionId_;
  String rtspURL_;
  Client rtspClient_;

  VoyContext context_;

  VOY_Database db_;

  int rtspSessionId_;

  public RecordController(VoyContext context)
  {
    context_ = context;
    
    addresses_ = null;
    sessionName_ = null;
    sessionDescription_ = null;
    rtspSessionId_ = -1;
    db_ = context_.getDB();
  }

  public void setAutomaticAddress()
  {
    System.out.println("RC: auto address");
    addressMode_ = AUTOMATIC;

    destinations_ = null;
    audioDestination_ = new Destination(null, 0);
    videoDestination_ = new Destination(null, 0);
  }

  public void setManualAddress(String addr)
  {
    System.out.println("RC: manual address: " + addr);

    addressMode_ = MANUAL;
    parseManualAddresses(addr);
  }

  public void setFLAddress( )
  {
    System.out.println("RecordController::setFLAddress");

    addressMode_ = FLBONE;
    String addr = "224.1.2.3/12000 224.1.2.3/12002";
    parseManualAddresses(addr);
  }

  public void setActinidesAddress( )
  {
    System.out.println("RecordController::setActinidesAddress");

    addressMode_ = ACTINIDES;
    String addr = "224.2.244.210/62670 224.2.241.69/29902";
    parseManualAddresses(addr);
  }
    
  public void setSessionInfo(String name, String desc)
  {
    System.out.println("Session info: " + name + " " + desc);

    sessionName_ = name;
    sessionDescription_ = desc;
  }

  public void commit() throws RTSPFailedException
  {
    System.out.println("commit!");

    rtspURL_ = createSession();

    setupRtsp();
  }

  String createSession()
  {
    String myhost;

    try {
      myhost = java.net.InetAddress.getLocalHost().getHostAddress();
    }
    catch (UnknownHostException e)
      {
	myhost = "unknown";
      }

    sessionId_ = db_.allocateRecordSession(sessionName_, sessionDescription_, myhost);

    System.out.println("Allocate returns id=" + sessionId_);

    int rc = db_.allocateRecordSpace(sessionId_);
	
    System.out.println("Allocate space returns rc=" + rc);

    return(db_.getURL(sessionId_));
  }

  //
  // Parse the address list and fill in the destinations_ array
  //
  void parseManualAddresses(String addr)
  {
    StringTokenizer tok = new StringTokenizer(addr);
    Vector vec = new Vector();

    while (tok.hasMoreTokens())
      {
	Destination d = null;
	String s = tok.nextToken();

	int slash = s.indexOf("/");
	if (slash >= 0)
	  {
	    String host = s.substring(0, slash);
	    String port = s.substring(slash + 1);
	    System.out.println("Got host=" + host + " port=" + port);
	    try {
	      d = new Destination(host, Integer.parseInt(port));
	    }
	    catch (NumberFormatException ex)
	      {
		System.out.println("invalid address");
	      }
	    if (d != null)
	      {
		vec.addElement(d);
	      }
		    
	  }
	else
	  {
	    //
	    // See if it's just a port
	    //
	    try {
	      int port = Integer.parseInt(s);
	      d = new Destination(null, port);
	    }
	    catch (NumberFormatException ex)
	      {
		System.out.println("invalid address");
	      }

	    if (d != null)
	      {
		vec.addElement(d);
	      }
	  }
      }

    destinations_ = new Destination[vec.size()];
    int i;
    for (i = 0; i < vec.size(); i++)
      {
	destinations_[i] = (Destination) vec.elementAt(i);
	System.out.println("Address " + i + " is " + destinations_[i]);
      }
  }

  //
  // Allocate the recording addresses and invoke the
  // RTSP SETUP calls
  //
  void setupRtsp() throws RTSPFailedException
  {
    rtspClient_ = new TCPClient();
	
    ConnectResult rc = rtspClient_.connect(rtspURL_);

    if (!rc.success())
      {
	throw(new RTSPFailedException("RTSP Connection to " + rtspURL_ + " Failed"));
      }

    //
    // Okay, we're ready to set up.
    //
    if (addressMode_ == AUTOMATIC)
      {
	setupForAutomatic();
      }
    else
      {
	setupForManual();
      }
  }

  void setupForManual() throws RTSPFailedException
  {
    int i;
    for (i = 0; i < destinations_.length; i++)
      {
	setupStream(destinations_[i], "stream_" + i);
      }
  }

  void setupForAutomatic() throws RTSPFailedException
  {
    setupStream(audioDestination_, "audio");
    setupStream(videoDestination_, "video");
  }

  //
  // Do the RTSP setup for this destination and stream name
  //
  void setupStream(Destination dest, String streamName) throws RTSPFailedException
  {
    SetupStreamResult rc;

    System.out.println("Setting up " + streamName + " to " + dest);
    rc = rtspClient_.setupStream(String.valueOf(sessionId_), streamName,
				 dest.transportString("RECORD"),
				 rtspSessionId_);
    System.out.println("Setup returns " + rc);

    if( !rc.success( ) )
	throw( new RTSPFailedException( "Setup failure" ) );

    if( rtspSessionId_ == -1 )
    {
	rtspSessionId_ = rc.rtspSessionId( );
	System.out.println( "Got a new sessionId " + rtspSessionId_ );
    }
    else
	System.out.println( "Using existing sessionId " + rtspSessionId_ );
  }

  //
  // Send rtsp RECORD request
  //
  public void recordSession() throws RTSPFailedException
  {
   RecordSessionResult rc;

    System.out.println( "Starting record " );
    rc = rtspClient_.recordSession( String.valueOf(sessionId_) ,  rtspSessionId_);

    if( !rc.success( ) )
	throw( new RTSPFailedException( "Recording failure" ) );

    System.out.println("Record returns " + rc);
  }

    public void teardownSession( )
    {
	int i;
	for (i = 0; i < destinations_.length; i++)
	{
	    teardownStream(destinations_[i], "stream_" + i);
	}
    }

    void teardownStream(Destination dest, String streamName)
    {
	TeardownStreamResult rc;

	System.out.println("Tearing down " + streamName + " to " + dest);
	rc = rtspClient_.teardownStream(String.valueOf(sessionId_), streamName,
				     rtspSessionId_);
	System.out.println("Teardown returns " + rc);
    }

}
