// $Id: RTSP_Request_Packet.C,v 1.2.2.1 1999/03/24 21:40:27 childers Exp $

#include "ace/Message_Block.h"

#include "RTSP_Request_Packet.h"
#include "HTTP_Helpers.h"

const char * const
RTSP_Request_Packet::static_header_strings_[RTSP_Request_Packet::NUM_HEADER_STRINGS]
= { "Date", "Pragma",
    "Authorization", "From", "If-Modified-Since", "Referrer", "User-Agent",
    "Allow", "Content-Encoding", "Content-Length", "Content-Type",
    "Expires", "Last-Modified", "Transport", "Session" };

const char * const
RTSP_Request_Packet::static_method_strings_[RTSP_Request_Packet::NUM_METHOD_STRINGS]
= { "DESCRIBE", "GET_PARAMETER", "OPTIONS", "PAUSE", "PLAY", "RECORD",
    "REDIRECT", "SETUP", "SET_PARAMETER", "TEARDOWN" };

// for reasons of efficiency, this class expects buffer to be
// null-terminated, and buflen does NOT include the \0

RTSP_Request_Packet::RTSP_Request_Packet (RTSP_Engine *engine)
  : got_request_line_ (0),
    method_ (0),
    uri_ (0),
    version_ (0),
    path_ (0),
    query_string_ (0),
    path_info_ (0),
    header_strings_ (RTSP_Request_Packet::static_header_strings_),
    method_strings_ (RTSP_Request_Packet::static_method_strings_),
    engine_(engine)
{
  for (int i = 0; i < RTSP_Request_Packet::NUM_HEADER_STRINGS; i++)
    this->headers_.recognize (this->header_strings_[i]);
}

RTSP_Request_Packet::~RTSP_Request_Packet (void)
{
  ACE_OS::free (this->method_);
  ACE_OS::free (this->uri_);
  ACE_OS::free (this->version_);
  ACE_OS::free (this->path_);
  ACE_OS::free (this->query_string_);
  ACE_OS::free (this->path_info_);


}

int
RTSP_Request_Packet::parse_request (ACE_Message_Block &mb)
{
    mb.wr_ptr ()[0] = '\0';

  // Note that RFC 822 does not mention the maximum length of a header
  // line.  So in theory, there is no maximum length.

  // In Apache, they assume that each header line should not exceed 8K.

//    printf("buf: %x %x %s\n", mb.rd_ptr(), mb.wr_ptr(), mb.rd_ptr());
    int result = this->headers_.complete_header_line (mb.rd_ptr ());
//    printf("Complete_header returns %d\n", result);
    if (result != 0)
    {
//	printf("got_request=%d\n", got_request_line());
	if (!this->got_request_line ())
        {
	    this->parse_request_line (mb.rd_ptr ());
	    while (this->headers_.complete_header_line (mb.rd_ptr ()) > 0)
		this->headers_.parse_header_line (mb.rd_ptr ());
        }
	else if (result > 0)
	{
	    do
	    {
		this->headers_.parse_header_line (mb.rd_ptr ());
	    }
	    while (this->headers_.complete_header_line (mb.rd_ptr ()) > 0);
	}
    }

    mb.wr_ptr (strlen(mb.rd_ptr ()) - mb.length ());

//     if (this->headers_.end_of_headers())
//     {
//	printf("Got eoh\n");
//     }
//     else if (this->got_request_line() && this->version() == 0)
//     {
//	printf("got req and vers = 0\n");
//     }
//    printf("rd_ptr=%x wr_ptr=%x length=%d\n",
// 	   mb.rd_ptr(), mb.wr_ptr(), mb.length());

    return ((this->headers_.end_of_headers ()
	     || (this->got_request_line () && this->version () == 0))
	    ? this->init (mb.rd_ptr (), mb.length ())
	    : 0);
}

void
RTSP_Request_Packet::parse_request_line (char * const request_line)
{
  char *ptr = request_line;
  char *buf = request_line;
  int offset = 1;

  this->status_ = HTTP_Status_Code::STATUS_OK;

//  printf("parse_request_line %s\n", request_line);

  ptr = ACE_OS::strchr (request_line, '\n');
  if (ptr > request_line && ptr[-1] == '\r')
    ptr--, offset++;
  if (ptr == request_line)
    {
      this->status_ = HTTP_Status_Code::STATUS_BAD_REQUEST;
      return;
    }
  *ptr = '\0';
  ptr += offset;

  char *lasts; // for strtok_r

  // Get the request type.
  this->got_request_line_ = 1;

  if (this->method (ACE_OS::strtok_r (buf, " \t", &lasts))
      && this->uri (ACE_OS::strtok_r (NULL, " \t", &lasts))
      && this->version (ACE_OS::strtok_r (NULL, " \t", &lasts))
      && this->sequence (ACE_OS::strtok_r (NULL, " \t", &lasts)))
    {
      this->type (this->method ());
      this->path(this->uri());
//      printf("parsed req line, type=%d\n", type());
    }

//   ACE_DEBUG ((LM_DEBUG, " (%t) request %s %s %s parsed\n",
//               (this->method () ? this->method () : "-"),
//               (this->uri () ? this->uri () : "="),
//               (this->version () ? this->version () : "HTTP/0.9")));

  ACE_OS::memmove (buf, ptr, ACE_OS::strlen (ptr)+1);
}

int
RTSP_Request_Packet::init (char * const buffer, int buflen)
{
  // Initialize these every time.
  content_length_ = -1;

  // Extract the data pointer.
  data_ = buffer;
  datalen_ = 0;

  // Set the datalen
  if (data_ != NULL)
    datalen_ = buflen;
  else
    datalen_ = 0;

//  ACE_DEBUG ((LM_DEBUG, " (%t) init has initialized\n"));

  return 1;
}

const char *
RTSP_Request_Packet::method (void) const
{
  return this->method_;
}

char *
RTSP_Request_Packet::uri (void) const
{
  return this->uri_;
}

const char *
RTSP_Request_Packet::sequence (void) const
{
  return this->sequence_;
}

const char *
RTSP_Request_Packet::version (void) const
{
  return this->version_;
}

const char *
RTSP_Request_Packet::path (void) const
{
  return this->path_;
}


const char *
RTSP_Request_Packet::query_string (void) const
{
  return this->query_string_;
}

const char *
RTSP_Request_Packet::path_info (void) const
{
  return this->path_info_;
}

int RTSP_Request_Packet::got_request_line (void) const
{
  return this->got_request_line_;
}

int
RTSP_Request_Packet::type (void) const
{
  return type_;
}

const Headers &
RTSP_Request_Packet::headers (void) const
{
  return this->headers_;
}

const char *
RTSP_Request_Packet::header_strings (int index) const
{
  const char *hs = 0;
  if (0 <= index && index < NUM_HEADER_STRINGS)
    {
      hs = this->header_strings_[index];
    }

  return hs;
}

const char *
RTSP_Request_Packet::header_values (int index) const
{
  const char *hs = 0;
  const char *hv = 0;

  if (0 <= index && index < NUM_HEADER_STRINGS)
    {
      hs = this->header_strings_[index];
      hv = this->headers_[hs];
    }

  return hv;
}

char *
RTSP_Request_Packet::data (void)
{
  return data_;
}

int
RTSP_Request_Packet::data_length (void)
{
  return datalen_;
}

int
RTSP_Request_Packet::content_length (void)
{
  if (this->content_length_ == -1)
    {
      const char * clv = this->headers_["Content-length"];
      this->content_length_ = (clv ? ACE_OS::atoi (clv) : 0);
    }
  return this->content_length_;
}

int
RTSP_Request_Packet::status (void)
{
  return this->status_;
}

const char *
RTSP_Request_Packet::status_string (void)
{
  return HTTP_Status_Code::instance ()[this->status_];
}

void
RTSP_Request_Packet::dump (void)
{
  ACE_DEBUG ((LM_DEBUG, "%s command.\n"
              "filename is %s,"
              " length of the file is %d,"
              " data string is %s,"
              " datalen is %d,"
              " status is %d, which is %s\n\n", 
              this->method () ? this->method () : "EMPTY",
              this->uri () ? this->uri () : "EMPTY",
              this->content_length (),
              this->data () ? this->data () : "EMPTY",
              this->data_length (),
              this->status (),
              this->status_string ()));
}

const char *
RTSP_Request_Packet::method (const char *method_string)
{
  if (this->method_) ACE_OS::free (this->method_);

  if (method_string == 0)
    {
      this->status_ = HTTP_Status_Code::STATUS_BAD_REQUEST;
      this->method_ = 0;
    }
  else this->method_ = ACE_OS::strdup (method_string);

  return this->method_;
}

const char *
RTSP_Request_Packet::uri (char *uri_string)
{
  if (this->uri_) ACE_OS::free (this->uri_);


  if (uri_string == 0)
  {
      this->status_ = HTTP_Status_Code::STATUS_BAD_REQUEST;
      this->uri_ = 0;
  }
  else
  {
      this->uri_ =  ACE_OS::strdup (uri_string);
      HTTP_Helper::HTTP_decode_string (this->uri_);
  }
  
  return this->uri_;
}

const char *
RTSP_Request_Packet::sequence (char *sequence_string)
{
  if (this->sequence_) ACE_OS::free (this->sequence_);

  if (sequence_string == 0)
    {
      this->sequence_ = 0;
    }
  else
    {
      this->sequence_ =  ACE_OS::strdup (sequence_string);
    }

  return this->sequence_;
}

const char *
RTSP_Request_Packet::version (const char *version_string)
{
  if (this->version_) ACE_OS::free (this->version_);

  this->version_ = (version_string
                    ? ACE_OS::strdup(version_string)
                    : 0);

  return this->version_;
}

int
RTSP_Request_Packet::type (const char *type_string)
{
  this->type_ = RTSP_Request_Packet::NO_TYPE;
  if (type_string == 0) return this->type_;

  for (int i = 0; i < RTSP_Request_Packet::NUM_METHOD_STRINGS; i++)
    if (ACE_OS::strcmp(type_string, this->method_strings_[i]) == 0)
      {
        this->type_ = i;
        break;
      }

  if (this->type_ == RTSP_Request_Packet::NO_TYPE)
    this->status_ = HTTP_Status_Code::STATUS_NOT_IMPLEMENTED;

  return this->type_;
}

const char *
RTSP_Request_Packet::path (char *uri_string)
//RTSP_Request_Packet::path (const char *uri_string) CHANGED FOR NEW ACE
{
    if (this->path_)
    {
	ACE_OS::free(this->path_);
	this->path_ = 0;
    }
    //
    // Parse out the abs/relative stuff in the url
    //
    
    if (strncmp(uri_string, "rtsp:", 5) == 0)
    {
//	printf("Schemed URI\n");

	uri_string += 5;
    }

    if (strncmp(uri_string, "//", 2) == 0)
    {
//	printf("host info\n");
	uri_string += 2;
	//const char *hostp = uri_string; CHANGED FOR NEW ACE
	char* hostp = uri_string;
	char host[1024];
	char *slash_idx = ACE_OS::strchr(hostp, '/');
	if (slash_idx)
	{
	    strncpy(host, hostp, slash_idx - hostp);
	    host[slash_idx - hostp] = '\0';
	    uri_string = slash_idx;
	}
	else
	{
	    strcpy(host, hostp);
	}
//	printf("Have host info '%s'\n", host);
    }

    if (uri_string[0] != '/')
    {
	//printf("Warning: Relative urls not allowed\n");
    }
    else
    {
//	printf("absolute uri\n");
	uri_string++;
    }
    
//    printf("Uri: '%s'\n", uri_string);
    
    this->path_ =  ACE_OS::strdup (uri_string);
    return this->path_;
}
