#include "Record_Session_Manager.h"
#include "RTSP_Setup_Request.h"
#include "RTSP_Record_Request.h"
#include "Record_Network_Transport.h"
#include "RTSP_Response.h"
#include "RTSP_Request_Packet.h"
#include "Record_Listener.h"
#include "Record_Archive_Manager.h"
#include "ace/OS.h"
#include "ace/Service_Config.h"
#include "ace/Reactor.h"
#include "RTSP_Teardown_Request.h"


Record_Session_Manager::Record_Session_Manager( VOY_Database_ptr db ,
						Record_Session_ID id ,
						RTSP_Setup_Request *request ) :
    databaseRef_( db ) ,   sessionId_(id)
{
    char host[ 256 ];
    ::gethostname( host , 256 );
    
    dbSessionId_ = request->sessionId( );

   printf( "Record_Session_Manager::Record_Session_Manager: Getting session path, host %s, rtsp id %d, db id %d\n" , host ,(int) id , dbSessionId_ );

    CORBA_String_var sessionPath = db->getSessionPath( dbSessionId_ ,
						       CORBA_string_dup( host ) );

    char* directory = ACE_OS::strdup( sessionPath);
    
    printf( "Record_Session_Manager::Record_Session_Manager: got path %s\n" , directory );
    
    archiveMgr_ = new Record_Archive_Manager( directory , dbSessionId_ , id , db );

    assert( archiveMgr_ );
}

Record_Session_Manager::~Record_Session_Manager( )
{
    delete archiveMgr_;
}

//
// Add a stream to the recording. Note that stream in this sense
// is a logical stream, matching a particular network host/port pair.
// Each logical stream may generate many actual RTP streams, one for
// each sender on that channel (in the multicast case).
// We create a Record_Streamer to sink each of the logical streams.
//
// The streamer is responsible for opening up the destination file
// on the media filesystem (it makes db lookups to find the path
// it should use).
//
Record_Session_Manager::RecordSessionError
Record_Session_Manager::addSender(RTSP_Setup_Request* sRequest,
				  Record_Network_Transport* netTrans)
{
    assert( sRequest && netTrans && archiveMgr_ );
    
    RecordSessionError errCode = kFailure;
    
    cout << "Record_Session_Manager:  Adding listener for setup " << *sRequest << endl;

    ACE_Reactor* pReactor = ACE_Reactor::instance( );
    Record_Listener* listener = new Record_Listener( this , netTrans , archiveMgr_ , pReactor );
    
    if( listener )
    {
	char streamIdc[ 256 ];
	sRequest->streamId( streamIdc );
	VOY_String streamId( streamIdc );
	printf( "Record_Session_Manager::addSender %u\n" , listener );
	listenMap_[ streamId ] = listener;

	RTSP_Request_Packet *request = sRequest->requestPacket( );
	RTSP_Response resp( 200 , "OK" , atoi( request->sequence( ) ) );
	char sessionId[ 256 ];
	sprintf( sessionId , "%d" , (int) sessionId_ );
	resp.setHeader( "Session" , sessionId );
	resp.send( request->engine( ) );

	errCode = kSuccess;
    }

    return errCode;
}

Record_Session_Manager::RecordSessionError
Record_Session_Manager::record( RTSP_Record_Request *request )
{
    RecordSessionError errCode = kSuccess;

    archiveMgr_->record( request );

    ListenMap::iterator iter;
    for( iter = listenMap_.begin( ) ; iter != listenMap_.end( ) ; iter++ )
    {
	printf( "Record_Session_Manager::record: calling record( ) for %u\n" , (*iter).second );
	(*iter).second->record( request );
    }

    return errCode;
}

void
Record_Session_Manager::teardown( RTSP_Teardown_Request* tRequest ,
				  Transport_Manager* transMan ,
				  short& qNoMoreStreams )
{
    char streamIdc[ 256 ];
    tRequest->streamId( streamIdc );
    VOY_String streamId( streamIdc );

    printf( "Record_Session_Manager::teardown: tearing down stream %s\n" , streamId.c_str( ) );

   ListenMap::iterator pListener = listenMap_.find( streamId );

    if( pListener != listenMap_.end( ) )
    {
 	printf( "Record_Session_Manager::teardown: GOOD TEARDOWN for Listener %u!\n" ,
		(*pListener).second );
	
	// Pass the teardown request to the guy listening on the socket
	(*pListener).second->teardown( transMan );

	// Clean up time
	//
	delete (*pListener).second;
	listenMap_.erase( streamId );
	
	// Send a response to the teardown request
	RTSP_Request_Packet *request = tRequest->requestPacket( );
	RTSP_Response resp( 200 , "OK" , atoi( request->sequence( ) ) );
	char sessionId[ 256 ];
	sprintf( sessionId , "%d" , (int) sessionId_ );
	resp.setHeader( "Session" , sessionId );
	resp.send( request->engine( ) );
    }
    else
    {
	// Send a bad response
	printf( "Record_Session_Manager::teardown: BAD TEARDOWN!\n" );
	assert( 0 );
     }

    // If there are no more streams, the parent Record_Manager will delete
    // this Record_Session_Manager
    if( (qNoMoreStreams = listenMap_.empty( )) )
    {
	// Kill the objects waiting to write packets to disk
	archiveMgr_->terminate( );
    }
}

