#define USE_STDIO_BUFFERING

#define NO_MMFS

#ifndef _AIX
#define NO_MMFS
#endif

#ifndef NO_MMFS
#include "TigerShark.h"
#endif

#include "Stream_Source.h"
#include "stdio.h"
#include "Play_Stream_Synchronizer.h"
#include "Playback_Network_Transport.h"
#include "RTSP_Setup_Request.h"
#include "RTSP_Play_Request.h"
#include "ace/Reactor.h"
#include "ace/OS.h"
#include <sys/stat.h>
#include "limits.h"
#include "config.h"

extern double gettimeofday();

/////////////////////////////////////////////////////////////////////////////
//
// Stream_Source.C
//
// Revision History
//     17 Sep 97 - lisa - original version
//


Stream_Source::Stream_Source( VOY_String payFreqFileName ) : synch_( NULL ) , 
    netTrans_( NULL ) , streamFile_( NULL ) , fileBuffer_( NULL ) , 
    lastPacketTimestamp_( 0 ) , qEof_( 0 ) , wrapAroundTimestampCount_( 0 )
{
    init( payFreqFileName );
}

void
Stream_Source::init( VOY_String payFreqFileName )
{
    streamSpec_[ 0 ] = '\0';

    // Load the default payload frequencies
    payloadFreqMap_[ 0 ] = 8000;  // PCMU
    payloadFreqMap_[ 1 ] = 8000;  // 1016
    payloadFreqMap_[ 2 ] = 8000;  // G721
    payloadFreqMap_[ 3 ] = 8000;  // GSM
    payloadFreqMap_[ 4 ] = 8000;  // unassigned
    payloadFreqMap_[ 5 ] = 8000;  // DVI4
    payloadFreqMap_[ 7 ] = 8000;  // LPC
    payloadFreqMap_[ 8 ] = 8000;  // PCMA
    payloadFreqMap_[ 9 ] = 8000;  // G722
    payloadFreqMap_[ 15 ] = 8000; // G728
    payloadFreqMap_[ 6 ] = 16000; // DVI4
    payloadFreqMap_[ 10 ] = 44100; // L16 stereo
    payloadFreqMap_[ 11 ] = 44100; // L16 mono
    payloadFreqMap_[ 14 ] = 90000; // MPA
    payloadFreqMap_[ 25 ] = 90000; // CelB
    payloadFreqMap_[ 26 ] = 90000; // JPEG
    payloadFreqMap_[ 28 ] = 90000; // nv
    payloadFreqMap_[ 31 ] = 90000; // H261
    payloadFreqMap_[ 32 ] = 90000; // MPV
    payloadFreqMap_[ 33 ] = 90000; // MPA

    // try to load in the file data
    if( payFreqFileName.length( ) > 0 )
    {
	ifstream payFreqFile( payFreqFileName.c_str( ) , ios::in );
	if( !payFreqFile )
	{
	    printf( "Stream_Source::Stream_Source:  Unable to open the payload-frequency data file %s\n" , payFreqFileName.c_str( ) );
	    exit( 1 );
	}
	else
	{
	    long freq;
	    int payload;

	    payFreqFile >> payload >> freq;
	    while( !payFreqFile.eof( ) && !payFreqFile.fail( ) )
	    {
		printf( "Stream_Source::Stream_Source:  just got payload %d, freq %ld\n" , payload , freq );
		payloadFreqMap_[ payload ] = freq;
		payFreqFile >> payload >> freq;
	    }
	    //printf("Stream_Source::Stream_Source: finished reading frequencies\n" );
	}
    }
}

Stream_Source::~Stream_Source( void )
{
    teardown( );
    
    if( streamFile_ )
	fclose( streamFile_ );
}

void Stream_Source::setup(   VOY_Database_ptr db ,
			   RTSP_Setup_Request* request ,
			   Play_Stream_Synchronizer* synch ,
			   Playback_Network_Transport* netTrans )
{
    assert( synch && netTrans && db && request );
    
    synch_ = synch;
    netTrans_ = netTrans;

    short qSuccess = 0;

    // Open a connection to the data and process some meta-info
    if( openStream( db , request ) )
    {
	if( readHeader( ) )
	{
	    if( readInfoFile( ) )
	    {
		if( calcTime( ) )
		{
		    qSuccess = 1;
		}
	    }
	}
    }

    if( qSuccess && synch_  )
    {
	synch_->registerStream( this , startTime_  );
    }
    else
    {
	// throw exception
	assert( 0 );
    }
}

void Stream_Source::play( RTSP_Play_Request* request , ACE_Time_Value playbackStart ,
			  ACE_Time_Value baseTime )
{
    //printf( "Stream_Source::play: Got a play request for %s\n" , streamSpec_ );
    // Calculate new start time, so that we're aligned w/other streams in this session
    Nstart_ = (double) baseTime.sec() + (double) baseTime.usec() / 1e6;
    //printf( "Stream_Source::play:  Nstart=%f, Nsynch=%f, Rstart=%f, Rsynch=%f, F=%f\n", Nstart_ , Nsynch_ , Rstart_ , Rsynch_ , F_  );

    playbackStartD_ = (double) playbackStart.sec() +(double) playbackStart.usec() / 1000000.0;

    short qFoundRtpPacket = 0;
    while( !qFoundRtpPacket && !qEof_ )
    {
	short qError;
	readPacket( qFoundRtpPacket , qError , qEof_ );
	if( qError || qEof_ )   // An eof on our first packet is an error too
	{
	    terminateStream( );
	    assert( 0 );
	    // exception
	    return;
	}
    }
    //printf( "Stream_Source::play:  read the first RTP packet, initial timestamp is %u, freq is %u\n" , startingTimestamp_ ,  timestampFrequency_  );

    ACE_Time_Value rtcp_interval( 10 , 0 );

     // Start up the timer for the RTCP reports
    timerIds_[ kRTCP_Report ] = ACE_Reactor::instance( )->schedule_timer( this ,
		                       (void*)kRTCP_Report , 0 ,  rtcp_interval );

    double playoutDelay = computePlayoutDelay( );
    
    if( playoutDelay > 0.010 )
	scheduleTimer( playoutDelay );
    else
	sendCurrentAndScheduleNextPacket( );
}

// virtual
int Stream_Source::handle_timeout( const ACE_Time_Value& unused1 ,
					   const void* eventType )
{
    switch( (int)eventType )
    {
    case kCounter:
	printf( "Stream_Source::handle_timeout: kCounter event\n" );
	break;
    case kUpdate:
	//printf( "Stream_Source::handle_timeout: kUpdate event\n" );
	sendCurrentAndScheduleNextPacket( );
	break;
    case kRTCP_Report:
	//printf( "Stream_Source::handle_timeout: kRTCP_Report\n" );
	break;
    default:
	printf( "Stream_Source::handle_timeout: Invalid event\n" );
	assert( 0 );
	break;
    }
    return 0;
}

short Stream_Source::openStream( VOY_Database_ptr db , RTSP_Setup_Request* request )
{
    short qSuccess = 0;
    
    // Gather data needed to retrieve from the db the name of the file that contains
    // stream data
    int id = request->sessionId( );
    
    char host[ 256 ];
    ::gethostname( host, 256 );
    
    char streamId[ 256 ];
    request->streamId( streamId );
    
    //printf( "Stream_Source::openStream: host %s, sessionId %d, streamId %s\n", host , id , streamId );
    CORBA_String_var streamSpec = db->getStreamPath( id , CORBA_string_dup(host) , streamId );

    // Open the file for reading
    if( openStreamFile( streamSpec ) )
    {
	//printf( "Stream_Source::openStream: opened stream file\n" );
	 qSuccess = 1;
    }
    else
	 printf( "Stream_Source::openStream:  failed to open stream file\n" );

    return qSuccess;
}

short Stream_Source::openStreamFile( char* streamSpec )
{
    short qSuccess = 1;

    strcpy( streamSpec_ , streamSpec );

    printf("Stream Spec: <%s>\n", streamSpec);
    
    struct stat s;
    streamFile_ = fopen( streamSpec_  , "r"  );

    if( streamFile_ == (FILE *) NULL )
    {
	printf( "Stream_Source::openStreamFile: bad open %s\n" , streamSpec_  );
	qSuccess = 0;
	assert( 0 );
    }

    //printf( "Stream_Source::openStreamFile: Opened %s\n" , streamSpec_  );

    if( fstat( fileno( streamFile_  ) , &s ) < 0 )
    {
	printf( "Stream_Source::openStreamFile: bad fstat\n" );
	qSuccess = 0;
	assert( 0 );
    }

    //printf( "Stream_Source::openStreamFile: stat %s: blocksize=%d filesize=%d\n", 
    //streamSpec_, s.st_blksize , s.st_size );

#ifdef USE_STDIO_BUFFERING
    //printf( "Stream_Source::openStreamFile: Setting stdio blocksize to %d\n" , s.st_blksize );
    fileBuffer_ = new char[ s.st_blksize ];

    if( setvbuf( streamFile_ , fileBuffer_ ,  _IOFBF , s.st_blksize ) != 0 )
    {
	printf( "Stream_Source::openStreamFile: setvbuf problem\n" );
	qSuccess = 0;
    }
#else
    fileBuffer_ = NULL;
    //printf("Stream_Source::openStreamFile: Disabling stdio buffering\n");
    if( setvbuf( streamFile_ , NULL , _IONBF, 0 ) != 0 )
    {
	printf( "Stream_Source::openStreamFile: setvbuf problem\n" );
	qSuccess = 0;
    }
#endif

#ifndef NO_MMFS
    if( s.st_vfstype == VFSTYPE_MMFS )
    {
	int ival;

	ival = 5000000;

#ifdef OLD_MMFS

	if (mmfs_qioctl(fileno(streamFile_), SET_RATE, &ival) < 0)
	{
	    printf( "Stream_Source::openStreamFile: SET_RATE problem\n" );
	    qSuccess = 0;
	}

	if (mmfs_qioctl(fileno(streamFile_), GET_RATE, &ival) < 0)
	{
	    printf( "Stream_Source::openStreamFile: GET_RATE problem\n" );
	    qSuccess = 0;
	}
	else
	    printf("rate=%d\n", ival);

	ival = TS_BUFFERS;
	if (mmfs_qioctl(fileno(streamFile_), SET_BUFFERING, &ival) < 0)
	{
	    printf( "Stream_Source::openStreamFile: SET_BUFFERING problem\n" );
	    qSuccess = 0;
	}
#else
	if (tsfattr(fileno(streamFile_), SET_RATE, &ival, NULL) < 0)
	{
	    printf( "Stream_Source::openStreamFile: SET_RATE problem\n" );
	    qSuccess = 0;
	}

	if (tsfattr(fileno(streamFile_), GET_RATE, &ival, NULL) < 0)
	{
	    printf( "Stream_Source::openStreamFile: GET_RATE problem\n" );
	    qSuccess = 0;
	}
	else
	    printf("rate=%d\n", ival);

	ival = TS_BUFFERS;
	if (tsfattr(fileno(streamFile_), SET_BUFFERING, &ival, NULL) < 0)
	{
	    printf( "Stream_Source::openStreamFile: SET_BUFFERING problem\n" );
	    qSuccess = 0;
	}
#endif

    }
    else
    {
	printf( "Stream_Source::openStreamFile: Warning: %s not in MMFS\n" , streamSpec_ );
	qSuccess = 0;
    }
#endif

    return qSuccess;
}

short Stream_Source::readHeader( void )
{
    short qSuccess = 1;
    
    if( fread( (void *)&header_ , sizeof( header_ ) , 1 , streamFile_  ) != 1 )
    {
	printf("Stream_Source::readHeader: fread failed\n");
	qSuccess = 0;
    }
    //printf( "Stream_Source::readHeader: read header, size=%d\n" , sizeof( header_ ) );

    startTime_.sec( ntohl( header_.start_s ) );
    startTime_.usec( ntohl( header_.start_us ) );
	    
    originalSource_.set( ntohs( header_.port ) , ntohl( header_.source ) );

    time_t time = (time_t) startTime_.sec( );
    //printf("Stream_Source::readHeader: Session was started by %s/%d at %d.%06d %s",
    //   originalSource_.get_host_name( ) ,
    //   originalSource_.get_port_number( ) ,
    //   startTime_.sec( ), startTime_.usec( ) ,
    //   ctime( &time ) );

    return qSuccess;
}

short Stream_Source::readInfoFile( void )
{
    short qSuccess = 1;
    FILE *infoFile;
    char ipath[1024];

    sprintf( ipath , "%s_info" , streamSpec_ );
    infoFile = fopen( ipath , "r" );

    if( !infoFile )
    {
	printf("Stream_Source::readInfoFile: Couldn't open info file %s\n" , ipath );
	qSuccess = 0;
	assert( 0 );
    }

    char line[1024];
    char*s;
    char* t;
    int size = sizeof( line );
    while( fgets( line , size , infoFile ) != NULL && qSuccess )
    {
	s = strchr( line , ':' );
	if( s )
	{
	    *s = '\0';
	    s++;
	    while( isspace(*s) )
		s++;
	    t = s + 1;
	    while( !isspace(*t) )
		t++;
	    *t = '\0';

	    qSuccess = handleInfoItem( line , s );
	}
    }
    fclose( infoFile );
    return qSuccess;
}

short Stream_Source::handleInfoItem( char *tag , char *value )
{
    short qSuccess = 1;
    
    if( strcmp( tag , "starting_timestamp" ) == 0 )
    {
#ifdef _AIX
	startingTimestamp_ = strtoul(value, 0, 10);
#else
	startingTimestamp_ = atol(value);
#endif
	Rstart_ = (double) startingTimestamp_;
    }
    else if( strcmp( tag , "packet_types" ) == 0 )
    {
	if( !packetType( value ) )
	{
	    printf("Stream_Source::handleInfoItem: Error parsing packet type\n");
	    qSuccess = 0;
	    assert( 0 );
	}
    }
    else if( strcmp( tag , "synch_ntp_timestamp" ) == 0 )
    {
	Nsynch_ = synchNtpTimestamp_ = atof(value);
    }
    else if( strcmp( tag , "synch_rtp_timestamp" ) == 0 )
    {
#ifdef _AIX
	synchRtpTimestamp_ = strtoul(value, 0, 10);
#else
	synchRtpTimestamp_ = atol(value);
#endif
//	synchRtpTimestamp_ = atoi(value);
	Rsynch_ = (double) synchRtpTimestamp_;
    }
    //else
	//printf( "Stream_Source::handleInfoItem:  unhandled tag %s\n" , tag );

    return qSuccess;
}

short Stream_Source::packetType( char* value )
{
    short qSuccess = 1;
    
    if (strchr(value, ' '))
    {
	printf("Stream_Source::packetType: Cannot handle streams with multiple packet types\n");
	qSuccess = 0;
    }
    
    ptype_ = atoi(value);

    PayloadFreqMap::iterator payloadIter;
    payloadIter = payloadFreqMap_.find( ptype_ );

    if( payloadIter == payloadFreqMap_.end( ) )
    {
	printf("Stream_Source::packetType: unknown packet type %d\n", ptype_);
	qSuccess = 0;
    }
    else
        timestampFrequency_ = (*payloadIter).second;

    F_ = (double) timestampFrequency_;
    
    return qSuccess;
}

short Stream_Source::calcTime( void )
{
    short qSuccess = 1;
    
    double newStartTime = rtpToNtp( startingTimestamp_ );
    //printf("Stream_Source::calcTime: SSRC %s: Recalculating start_time: was %d.%06d, now %f\n", streamSpec_, startTime_.sec() , startTime_.usec() , newStartTime );
    
    long stsec = (int) newStartTime;
    startTime_.set(stsec, 1e6 * (newStartTime - (double) stsec));
    //printf("Stream_Source::calcTime: Recalculating start_time: now %d.%06d\n", startTime_.sec(), startTime_.usec());

    /*
     * Compute some of the synchronization constants.
     */

    Nstart_ = Nsynch_ - (Rsynch_ - Rstart_) / F_;
    //printf("Stream_Source::calcTime: Nstart=%f Nsynch=%f Rstart=%f Rsynch=%f F=%f\n", Nstart_, Nsynch_, Rstart_, Rsynch_, F_);

    return qSuccess;
}


/*
* Transform an RTP timestamp to the NTP timestamp.
* The returned timestamp is the absolute time corresponding to
* this timestamp.
*/
double Stream_Source::rtpToNtp( double Ri )
{
    return  (Ri - Rsynch_) / F_ + Nsynch_;
}


/*
* Transform an RTP timestamp to the NTP timestamp.
* The returned timestamp is the absolute time corresponding to
* this timestamp.
*/
double Stream_Source::rtpToNtp( unsigned long rtpTimestamp )
{
    double Ri, Ni;

    Ri = (double) rtpTimestamp;

    Ni = (Ri - Rsynch_) / F_ + Nsynch_;

    return Ni;
}


void Stream_Source::readPacket( short& qFoundRtpPacket , short& qError , short& qEof )
{
    qError = 0;
    qFoundRtpPacket = 0;
    qEof = 0;

    //printf( "Stream_Source::readPacket: reading a packet\n" );
    
    RTP_Packet::RTP_Header *rp;

    // Read packet header
    if( fread( (void *) &packet_ , sizeof(packet_) , 1 , streamFile_  )  != 1 )
    {
	if( feof( streamFile_ ) )
	{
	    printf("Stream_Source::readPacket: end of file for %s\n" , streamSpec_ );
	    qEof = 1;
//	    ACE_Reactor::instance( )->cancel_timer( timerIds_[ kRTCP_Report ] );
	}
	else
	{
	    printf("Stream_Source::readPacket: read error in header\n");
	    qError = 1;
	}
    }
    else
    {
	// adjust byte order
	packet_.length = ntohs(packet_.length);
	packet_.type = ntohs(packet_.type);
	packet_.offset = ntohl(packet_.offset);

	//printf( "Stream_Source::readPacket: pkt: len=%d type=%d offset=%d\n" ,
	//packet_.length , packet_.type , packet_.offset );

	// calculate packet body size and read it
	packetBufferLength_ = packet_.length - sizeof( packet_ );
	if( fread( packetBuffer_ , packetBufferLength_ , 1 , streamFile_  ) != 1 )
	{
	    fprintf( stderr, "Stream_Source::readPacket: fread body %d failed: %s\n",
		     packetBufferLength_ ,  strerror(errno));
	    qError = 1;
	}
	else
	{
	    //printf( "Stream_Source::readPacket: fread body %d succeeded\n" ,
	    //packetBufferLength_ );

	    if( packet_.type == 0 )
	    {
		rp = ( RTP_Packet::RTP_Header * ) packetBuffer_;
		packetTimestamp_ = rp->get_timestamp( );
		//printf( "Stream_Source::readPacket: got the RTP packet\n" );
		qFoundRtpPacket = 1;
	    }
	}
    }
}
	

void Stream_Source::sendCurrentAndScheduleNextPacket( void )
{
    //printf( "Stream_Source::sendCurrentAndScheduleNextPacket: start\n" );

    // Send out the packet that's waiting
    sendPacket();
    
    short qError = 0;
    short qTimerScheduled = 0;

    // Execute this loop until we've queued up the next packet
    while( !qTimerScheduled  &&  !qEof_  &&  !qError )
    {
	short qFoundRtpPacket;

	// Read the next packet
	readPacket( qFoundRtpPacket , qError , qEof_ );
	
	if( qError )
	{
	    terminateStream();
	    assert( 0 );
	    // exception
	}
	else if( !qEof_ ) // We read some data; look at what we have
	{
	    if( qFoundRtpPacket )
	    {
		if( packetTimestamp_ == lastPacketTimestamp_ )
		{
		    //printf( "Stream_Source::sendCurrentAndScheduleNextPacket: dup RTP packet %u %u\n" ,
		    //packetTimestamp_  , lastPacketTimestamp_ );
		    sendPacket();
		    continue;
		}
		else
		{
		    //printf( "Stream_Source::sendCurrentAndScheduleNextPacket: nondup RTP packet %u %u\n" ,
		    //packetTimestamp_  , lastPacketTimestamp_ );
		}

		double playoutDelay = computePlayoutDelay();
	    
		// Schedule a timer for sending the packet if the delay is
		// greater than 10ms.  This stops the loop.
		if (playoutDelay > 0.010)
		{
		    scheduleTimer( playoutDelay );
		    qTimerScheduled = 1;
		}
		else // send it right away.
		    sendPacket( );
	    }
	    else  // the data we read was an RTCP packet.  Just send it along
		sendPacket( );
	}
	else
	{
	    // Stream has finished.  Clean up
	    terminateStream( );
	}
    }
}

void Stream_Source::sendPacket( void )
{
    if( packet_.type == 0 )
    {

	lastPacketTimestamp_ = packetTimestamp_;
	//printf( "Stream_Source::sendPacket: sending an RTP packet\n" );
	RTP_Packet::RTP_Header *rp = (RTP_Packet::RTP_Header *)packetBuffer_;
	netTrans_->playRtpPacket( rp->get_packet_type( ) , packetBuffer_,
					      packetBufferLength_ );
    }
    else
    {
	//printf( "Stream_Source::sendPacket: sending a RTCP packet\n" );
	netTrans_->playRtcpPacket( packetBuffer_, packetBufferLength_ );
    }
}

void Stream_Source::terminateStream( void )
{
    //printf("Stream_Source::terminateStream %x done\n", this);

    // Not sure what should go here
}


/*
* Compute the playout delay for the packet currently buffered
* in packet_/packetBuffer_.
*/
double Stream_Source::computePlayoutDelay( void )
{
    double cur_d = gettimeofday( );

    double rel = rtpToRelNtp( packetTimestamp_  );
    double delay = rel + playbackStartD_ - cur_d;


    if( delay > 60.0 )
    {
	printf( "Stream_Source::computePlayoutDelay: Hey! BIG delay for %s\n" ,
		streamSpec_ );
	printf( "Stream_Source::computePlayoutDelay: ts=%u rel=%f start=%f cur=%f delay=%f\n" ,
		packetTimestamp_ , rel , playbackStartD_ , cur_d , delay );
    }
    else if( delay < 0.0 )
    {
	//printf( "Stream_Source::computePlayoutDelay: SSRC %s: Hey! NEGATIVE delay\n", streamSpec_);
	// assert( 0 );
	// exception ?
    }
    return delay;
}


/*
* Transform an RTP timestamp to the NTP timestamp.
* Timestamp returned is relative to the start of the recording.
*/
double Stream_Source::rtpToRelNtp(unsigned long rtp_ts)
{
    if( ( lastPacketTimestamp_ > rtp_ts )  &&
        ( ( lastPacketTimestamp_ - rtp_ts ) > ( ULONG_MAX / 10000 ) ) )
    {
	// Adjust rtp_ts values that have wrapped back around to the beginning of ULONG space
	wrapAroundTimestampCount_++;
	//printf( "Stream_Source::rtpToRelNtp: Timestamp wrap around!  count=%d\n" , wrapAroundTimestampCount_ );
    }

    double tsPlusWrap = (double)rtp_ts;

    if( wrapAroundTimestampCount_ > 0 )
    {
	tsPlusWrap += (double)( (double)wrapAroundTimestampCount_ * (double)ULONG_MAX);
    }
	
    double Ni = rtpToNtp( tsPlusWrap );
    Ni = Ni - Nstart_;

    return Ni;
}

void Stream_Source::scheduleTimer( double playoutDelay )
{
    long  s , us;

    //printf( "Stream_Source::scheduleTimer:  We're scheduling a time for an RTP packet\n" );
    
    s = (long)playoutDelay;
    
    us = (long)( ( playoutDelay - (double)s ) * 1000000.0 );
    
    timerIds_[ kUpdate ] = ACE_Reactor::instance( )->schedule_timer( this , (void *)kUpdate ,
						   ACE_Time_Value( s , us ) , 0 );
}

void Stream_Source::teardown( void )
{
    if( synch_ )
    {
	synch_->unregisterStream( this );
	synch_ = NULL;
    }

    cancelTimers( );
}

void Stream_Source::cancelTimers( void)
{
   ACE_Reactor::instance( )->cancel_timer( this );
}
