#include "Transport_Manager.h"
#include "Record_Network_Transport.h"
#include "Playback_Network_Transport.h"
#include "RTSP_Setup_Request.h"
#include "RTSP_Request_Packet.h"
#include "RTSP_Engine.h"
#include "RTSP_Transport_Info.h"

#include "random.h"

#include <ace/INET_Addr.h>
#include <ace/SString.h>

/////////////////////////////////////////////////////////////////////////////
//
// Transport_Manager.C
//
// Revision History
//     17 Sep 97 - lisa - original version
//

Transport_Manager::Transport_Manager( )
{
}

Transport_Manager::~Transport_Manager( )
{
}

Record_Network_Transport*
Transport_Manager::setupRecordTransport( RTSP_Setup_Request* request )
{
    Record_Network_Transport* trans = NULL;
    Network_Endpoint ePoint;
    short qContinue = 1;
    
    //cout << "Transport_Manager::setupRecordTransport: for " << *request << endl;

    computeEndpoint( request, ePoint );

    //cout << "Transport_Manager::setupRecordTransport: computed endpoint " << ePoint << endl;
    
#ifndef VGER_MULTICAST_ENABLED
    if( ePoint.isMulticast ( ) )
    {
	cout << "Transport_Manager::setupRecordTransport: multicast requested but not supported!" << endl;
	qContinue = 0;
    }
#endif

    if( qContinue )
    {
	// Only record if the port is not currently being used by record
        if( !currentRecPort( ePoint.port_ ) )
	{
	    //cout << "Transport_Manager::setupRecordTransport: creating new transport\n";
	    trans = new Record_Network_Transport( ePoint );
	    
	    if( trans && !trans->qError( ) )
		storeRecHostPort( ePoint , trans );
	    else
	    {
		if( trans )
		    delete trans;
		trans = NULL;
	    }
	}
    }
    
    //cout << flush;
    
    return trans;
}

//
// A transport target is identified by the remote hostname
// and port number for the UDP sockets. The port number is for the
// RTP data stream. The RTCP port is the RTP port + 1
//
Playback_Network_Transport*
Transport_Manager::setupPlaybackTransport(RTSP_Setup_Request* request)
{
    Playback_Network_Transport* trans = NULL;
    Network_Endpoint ePoint;

    //cout << "Transport_Manager::setupPlaybackTransport: for " << *request << endl;

    computeEndpoint( request , ePoint );

    //cout << "Transport_Manager::setupPlaybackTransport: computed endpoint " <<ePoint<< endl;
    
    // Continue only if the port is not already in use by record
    cout << "Checking for use in record" << endl;
    if( !currentRecPort( ePoint.port_ ) )
    {
	cout << "Not used in record.  Checking for use in play" << endl;
	if( currentPlayHostPort( ePoint ) )
	{
	  // what if ttl is different????????????
	  trans = getPlayTransport( ePoint );
	  cout << "Transport_Manager::setupPlaybackTransport: reusing transport in play map " << *trans << endl;
	}
	else
	{
	    cout << "Transport_Manager::setupPlaybackTransport: creating new transport\n";
	    trans = new Playback_Network_Transport( ePoint );
	    if( trans && !trans->qError( ) )
		storePlayHostPort(ePoint , trans );
	    else
	    {
		if( trans )
		    delete trans;
		trans = NULL;
	    }
	}
    }

    cout << flush;
    
    return trans;
}

void
Transport_Manager::computeEndpoint(RTSP_Setup_Request* request,
				       Network_Endpoint& ePoint)
{
    //
    // Figure out the client's hostname
    //
    const ACE_INET_Addr &clientAddr = request->requestPacket()->engine()->peerAddr();
    VOY_String clientHostname(clientAddr.get_host_name());

    //
    // Determine the port to use.
    // If port is not present, or passed in as 0, select a random port
    //
    
    VOY_String portParam;
    int port;
    if (request->transportInfo()->getParameter("port", portParam) < 0)
	port = 0;
    else
	port = atoi(portParam.c_str());

    //
    // Assign a random port. Make sure we're not using it ourselves.
    //
    if (port == 0)
    {
	short qPortInUse = 1;
	cout << "assigning random port\n";
	do {
	    port = getRandomPort();
	    if( !currentRecPort( port ) && !currentPlayPort( port ) )
		qPortInUse = 0;
	} while( qPortInUse );
    }

    //
    // Determine the host.
    // If destination is not present, and we're recording, use INADDR_ANY
    // If destination is not present, and we're playing, use the client's address
    //
    VOY_String host;

    if (request->transportInfo()->getParameter("destination", host) < 0)
    {
	if (request->isModePlay())
	  {
	    cout << "Assigning Host: " << clientHostname.c_str( ) << " Port: " << port << endl;
	    ePoint.set(clientHostname, port);
	  }
	else if (request->isModeRecord())
	  {
	    ePoint.set(port);
	  }
    }
    else
      {
	ePoint.set(host, port);
      }

    // Look for ttl
    VOY_String ttlParam;
    int ttl;
    if (request->transportInfo()->getParameter("ttl", ttlParam) < 0)
	ttl = 16;
    else
	ttl = atoi(ttlParam.c_str());
    ePoint.ttl_ = ttl;
}

void
Transport_Manager::closeTransport( Record_Network_Transport* netTrans )
{
    Network_Endpoint ePoint = netTrans->endpoint( );

    removeRecHostPort( ePoint );
    delete netTrans;
}

short
Transport_Manager::currentRecHostPort( Network_Endpoint& ePoint )
{
  short isCurrent = 0;

  RecPortHostMap::iterator iterPort = recPortHostMap_.find( ePoint.port_ );
  if( iterPort != recPortHostMap_.end( ) )
  {
    RecHostMap* hostMap = (*iterPort).second;
    RecHostMap::iterator iterHost = hostMap->find( ePoint.host_ );
    if( iterHost != hostMap->end( ) )
      isCurrent = 1;
  }

  //printf( "Transport_Manager::currentRecHostPort:  found=%d, looking for port %d, host %s\n" , isCurrent, ePoint.port_ , ePoint.host_.c_str( ) );

  return isCurrent;
}

short
Transport_Manager::currentRecPort( int port )
{
    short isCurrent = 0;
    isCurrent = recPortHostMap_.find(port) != recPortHostMap_.end() ? 1 : 0;
    //printf( "Transport_Manager::currentRecPort: port %d found=%d\n", port , isCurrent );
    return isCurrent;
}

short
Transport_Manager::currentPlayPort( int port )
{
  return ( playPortHostMap_.find( port ) != playPortHostMap_.end( ) ) ? 1 : 0 ;
}

void
Transport_Manager::storeRecHostPort( Network_Endpoint& ePoint , Record_Network_Transport* trans )
{
  RecHostMap* hostMap = NULL;

  RecPortHostMap::iterator iter = recPortHostMap_.find( ePoint.port_ );
  if( iter != recPortHostMap_.end( ) )
  {
    hostMap = (*iter).second;
  }
  else
  {
    hostMap = new RecHostMap( );
    recPortHostMap_[ ePoint.port_ ] = hostMap;
  }

  (*hostMap)[ ePoint.host_ ] = trans;

  //printf( "Transport_Manager::storeRecHostPort: storing port %d, host %s\n" , ePoint.port_ , ePoint.host_.c_str( ) );
}

void
Transport_Manager::removeRecHostPort( Network_Endpoint& ePoint )
{
  RecHostMap* hostMap = NULL;

  RecPortHostMap::iterator iter = recPortHostMap_.find( ePoint.port_ );
  if( iter != recPortHostMap_.end( ) )
  {
    hostMap = (*iter).second;
    hostMap->erase( ePoint.host_ );
  }

  //printf( "Transport_Manager::removeRecHostPort: removing port %d, host %s\n" , ePoint.port_ , ePoint.host_.c_str( ) );
}

void
Transport_Manager::storePlayHostPort( Network_Endpoint& ePoint , Playback_Network_Transport* trans )
{
  PlayHostMap* hostMap = NULL;

  PlayPortHostMap::iterator iter = playPortHostMap_.find( ePoint.port_ );
  if( iter != playPortHostMap_.end( ) )
  {
    hostMap = (*iter).second;
  }
  else
  {
    hostMap = new PlayHostMap( );
    playPortHostMap_[ ePoint.port_ ] = hostMap;
  }

  (*hostMap)[ ePoint.host_ ] = trans;

  //printf( "Transport_Manager::storePlayHostPort: storing port %d, host %s\n" , ePoint.port_ , ePoint.host_.c_str( ) );
}

short
Transport_Manager::currentPlayHostPort( Network_Endpoint& ePoint )
{
  short isCurrent = 0;

  PlayPortHostMap::iterator iterPort = playPortHostMap_.find( ePoint.port_ );
  if( iterPort != playPortHostMap_.end( ) )
  {
    PlayHostMap* hostMap = (*iterPort).second;
    PlayHostMap::iterator iterHost = hostMap->find( ePoint.host_ );
    if( iterHost != hostMap->end( ) )
      isCurrent = 1;
  }

  //printf( "Transport_Manager::currentPlayHostPort:  found=%d, looking for port %d, host %s\n" , isCurrent, ePoint.port_ , ePoint.host_.c_str( ) );

  return isCurrent;
}

Playback_Network_Transport*
Transport_Manager::getPlayTransport( Network_Endpoint& ePoint )
{
  Playback_Network_Transport* trans = NULL;

  PlayPortHostMap::iterator iterPort = playPortHostMap_.find( ePoint.port_ );
  if( iterPort != playPortHostMap_.end( ) )
  {
    PlayHostMap* hostMap = (*iterPort).second;
    PlayHostMap::iterator iterHost = hostMap->find( ePoint.host_ );
    if( iterHost != hostMap->end( ) )
      trans = (*iterHost).second;
  }

  return trans;
}

Record_Network_Transport*
Transport_Manager::getRecTransport( Network_Endpoint& ePoint )
{
  Record_Network_Transport* trans = NULL;

  RecPortHostMap::iterator iterPort = recPortHostMap_.find( ePoint.port_ );
  if( iterPort != recPortHostMap_.end( ) )
  {
    RecHostMap* hostMap = (*iterPort).second;
    RecHostMap::iterator iterHost = hostMap->find( ePoint.host_ );
    if( iterHost != hostMap->end( ) )
      trans = (*iterHost).second;
  }

  return trans;
}
